import ctypes

from pyautd3.native_methods.autd3capi_driver import Duration as _Duration
from pyautd3.native_methods.autd3capi_driver import OptionDuration


class Duration:
    _inner: _Duration

    def __new__(cls: type["Duration"]) -> "Duration":
        raise NotImplementedError

    @classmethod
    def __private_new__(cls: type["Duration"], inner: _Duration) -> "Duration":
        ins = super().__new__(cls)
        ins._inner = inner
        return ins

    @staticmethod
    def from_nanos(ns: int) -> "Duration":
        inner = _Duration()
        inner.nanos = ns
        return Duration.__private_new__(inner)

    @staticmethod
    def from_micros(us: int) -> "Duration":
        return Duration.from_nanos(us * 1000)

    @staticmethod
    def from_millis(ms: int) -> "Duration":
        return Duration.from_micros(ms * 1000)

    @staticmethod
    def from_secs(s: int) -> "Duration":
        return Duration.from_millis(s * 1000)

    def as_nanos(self) -> int:
        return int(self._inner.nanos)

    def as_micros(self) -> int:
        return self.as_nanos() // 1000

    def as_millis(self) -> int:
        return self.as_micros() // 1000

    def as_secs(self) -> int:
        return self.as_millis() // 1000

    def __add__(self: "Duration", other: "Duration") -> "Duration":
        return Duration.from_nanos(self.as_nanos() + other.as_nanos())

    def __sub__(self: "Duration", other: "Duration") -> "Duration":
        return Duration.from_nanos(self.as_nanos() - other.as_nanos())

    def __mul__(self: "Duration", other: int) -> "Duration":
        return Duration.from_nanos(self.as_nanos() * other)

    def __truediv__(self: "Duration", other: int) -> "Duration":
        return Duration.from_nanos(self.as_nanos() // other)

    def __floordiv__(self: "Duration", other: int) -> "Duration":
        return Duration.from_nanos(self.as_nanos() // other)

    def __mod__(self: "Duration", other: int) -> "Duration":
        return Duration.from_nanos(self.as_nanos() % other)

    def __divmod__(self: "Duration", other: int) -> tuple[int, "Duration"]:
        quo, rem = divmod(self.as_nanos(), other)
        return quo, Duration.from_nanos(rem)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Duration):
            return False
        return self.as_nanos() == other.as_nanos()

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Duration):
            return True
        return self.as_nanos() != other.as_nanos()

    def __lt__(self, other: "Duration") -> bool:
        return self.as_nanos() < other.as_nanos()

    def __le__(self, other: "Duration") -> bool:
        return self.as_nanos() <= other.as_nanos()

    def __gt__(self, other: "Duration") -> bool:
        return self.as_nanos() > other.as_nanos()

    def __ge__(self, other: "Duration") -> bool:
        return self.as_nanos() >= other.as_nanos()

    def __hash__(self) -> int:
        return hash(self.as_nanos())

    def __str__(self) -> str:
        ns = self.as_nanos()
        if ns < 1000:  # noqa: PLR2004
            return f"{ns}ns"
        us = ns // 1000
        if us < 1000:  # noqa: PLR2004
            return f"{us}.{ns % 1000:03}".rstrip("0") + "μs" if ns % 1000 != 0 else f"{us}μs"
        ms = us // 1000
        if ms < 1000:  # noqa: PLR2004
            return f"{ms}.{us % 1000:03}".rstrip("0") + "ms" if us % 1000 != 0 else f"{ms}ms"
        s = ms // 1000
        return f"{s}.{ms % 1000:03}".rstrip("0") + "s" if ms % 1000 != 0 else f"{s}s"


def into_option_duration(duration: Duration | None) -> OptionDuration:
    return OptionDuration(ctypes.c_bool(duration is not None), duration._inner if duration is not None else _Duration())



import ctypes
from collections.abc import Iterable
from typing import Self

import numpy as np

from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.utils import _validate_nonzero_u32
from pyautd3.gain.holo.amplitude import Amplitude
from pyautd3.gain.holo.backend import Backend
from pyautd3.gain.holo.constraint import EmissionConstraint
from pyautd3.gain.holo.holo import HoloWithBackend
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.autd3capi_gain_holo import EmissionConstraintWrap
from pyautd3.native_methods.autd3capi_gain_holo import GSPATOption as GSPATOption_
from pyautd3.native_methods.structs import Point3


class GSPATOption:
    repeat: int
    constraint: EmissionConstraintWrap

    def __init__(self: Self, *, repeat: int = 100, constraint: EmissionConstraintWrap | None = None) -> None:
        self.repeat = _validate_nonzero_u32(repeat)
        self.constraint = constraint or EmissionConstraint.Clamp(EmitIntensity.MIN, EmitIntensity.MAX)

    def _inner(self: Self) -> GSPATOption_:
        return GSPATOption_(self.constraint, self.repeat)


class GSPAT(HoloWithBackend["GSPAT"]):
    option: GSPATOption

    def __init__(self: Self, backend: Backend, foci: Iterable[tuple[np.ndarray, Amplitude]], option: GSPATOption) -> None:
        super().__init__(foci, backend)
        self.option = option

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        size = len(self.foci)
        points = np.fromiter((np.void(Point3(d)) for d, _ in self.foci), dtype=Point3)  # type: ignore[type-var,call-overload]
        amps = np.fromiter((d.pascal() for _, d in self.foci), dtype=ctypes.c_float)  # type: ignore[type-var,call-overload]
        return self.backend._gspat(
            points.ctypes.data_as(ctypes.POINTER(Point3)),  # type: ignore[arg-type]
            amps.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            size,
            self.option._inner(),
        )



from .drive import Drive
from .emit_intensity import EmitIntensity
from .fpga_state import FPGAState
from .loop_behavior import LoopBehavior
from .phase import Phase
from .sampling_config import SamplingConfig
from .transition_mode import TransitionMode

__all__ = [
    "Drive",
    "EmitIntensity",
    "FPGAState",
    "LoopBehavior",
    "Phase",
    "SamplingConfig",
    "TransitionMode",
]



from enum import Enum
from typing import Generic, Self, TypeVar

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.utils import _validate_u8
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi import SquareOption as SquareOption_
from pyautd3.native_methods.autd3capi_driver import ModulationPtr

T = TypeVar("T", int, float)


class SquareOption:
    low: int
    high: int
    duty: float
    sampling_config: SamplingConfig

    def __init__(
        self: Self,
        *,
        low: int = 0x00,
        high: int = 0xFF,
        duty: float = 0.5,
        sampling_config: SamplingConfig | None = None,
    ) -> None:
        self.low = _validate_u8(low)
        self.high = _validate_u8(high)
        self.duty = duty
        self.sampling_config = sampling_config or SamplingConfig.FREQ_4K

    def _inner(self: Self) -> SquareOption_:
        return SquareOption_(
            self.low,
            self.high,
            self.duty,
            self.sampling_config.division,
        )


class SquareMode(Enum):
    Exact = 0
    ExactFloat = 1
    Nearest = 2


class Square(Modulation, Generic[T]):
    _mode: SquareMode
    freq: Freq[T]
    option: SquareOption

    def __init__(self: Self, freq: Freq[T], option: SquareOption) -> None:
        super().__init__()
        match freq.hz():
            case int():
                self._mode = SquareMode.Exact
            case _:
                self._mode = SquareMode.ExactFloat
        self.freq = freq
        self.option = option or SquareOption()

    def into_nearest(self: Self) -> "Square":
        match self._mode:
            case SquareMode.ExactFloat:
                new = Square(self.freq, self.option)
                new._mode = SquareMode.Nearest
                return new
            case _:
                raise TypeError

    def _modulation_ptr(self) -> ModulationPtr:
        match self._mode:
            case SquareMode.Exact:
                return Base().modulation_square_exact(self.freq.hz(), self.option._inner())  # type: ignore[arg-type]
            case SquareMode.ExactFloat:
                return Base().modulation_square_exact_float(self.freq.hz(), self.option._inner())
            case SquareMode.Nearest:  # pragma: no cover
                return Base().modulation_square_nearest(self.freq.hz(), self.option._inner())



import ctypes
from collections.abc import Callable
from typing import Self

from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga import Drive
from pyautd3.driver.geometry import Device, Geometry, Transducer
from pyautd3.native_methods.autd3 import Drive as Drive_
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ConstPtr, GainPtr, GeometryPtr


class Custom(Gain):
    def __init__(self: Self, f: Callable[[Device], Callable[[Transducer], Drive]]) -> None:
        super().__init__()

        def f_native(_context: ConstPtr, geometry_ptr: GeometryPtr, dev_idx: int, tr_idx: int, raw) -> None:  # noqa: ANN001
            dev = Device(dev_idx, geometry_ptr)
            tr = Transducer(tr_idx, dev_idx, dev._ptr)
            d = f(dev)(tr)
            raw[0] = Drive_(d.phase._inner(), d.intensity._inner())

        self._f_native = ctypes.CFUNCTYPE(None, ConstPtr, GeometryPtr, ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(Drive_))(f_native)

    def _gain_ptr(self: Self, geometry: Geometry) -> GainPtr:
        return Base().gain_custom(self._f_native, None, geometry._geometry_ptr)  # type: ignore[arg-type]



from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.defined.angle import Angle
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import BesselOption as BesselOption_
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Point3, Vector3


class BesselOption:
    intensity: EmitIntensity
    phase_offset: Phase

    def __init__(self: Self, *, intensity: EmitIntensity | None = None, phase_offset: Phase | None = None) -> None:
        self.intensity = intensity or EmitIntensity.MAX
        self.phase_offset = phase_offset or Phase(0)

    def _inner(self: Self) -> BesselOption_:
        return BesselOption_(self.intensity._inner(), self.phase_offset._inner())


class Bessel(Gain):
    pos: np.ndarray
    direction: np.ndarray
    theta: Angle
    option: BesselOption

    def __init__(self: Self, pos: ArrayLike, direction: ArrayLike, theta: Angle, option: BesselOption) -> None:
        super().__init__()
        self.pos = np.array(pos)
        self.direction = np.array(direction)
        self.theta = theta
        self.option = option

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        return Base().gain_bessel(
            Point3(self.pos),
            Vector3(self.direction),
            self.theta._inner(),
            self.option._inner(),
        )



def _validate_u8(value: int) -> int:
    if not isinstance(value, int):
        raise TypeError
    if value < 0 or value > 0xFF:  # noqa: PLR2004
        raise ValueError
    return value


def _validate_nonzero_u16(value: int) -> int:
    if not isinstance(value, int):
        raise TypeError
    if value <= 0 or value > 0xFFFF:  # noqa: PLR2004
        raise ValueError
    return value


def _validate_nonzero_u32(value: int) -> int:
    if not isinstance(value, int):
        raise TypeError
    if value <= 0 or value > 0xFFFFFFFF:  # noqa: PLR2004
        raise ValueError
    return value
